==> nrel.routee.compass-0.5.0.tar.gz: nrel.routee.compass-0.5.0/rust/routee-compass-powertrain/src/routee/vehicle/default/phev.rs <==
use std::sync::Arc;

use routee_compass_core::{
    model::traversal::{
        state::{state_variable::StateVar, traversal_state::TraversalState},
        traversal_model_error::TraversalModelError,
    },
    util::unit::{
        as_f64::AsF64, Distance, DistanceUnit, Energy, EnergyUnit, Grade, GradeUnit, Speed,
        SpeedUnit,
    },
};

use crate::routee::{
    prediction::PredictionModelRecord,
    vehicle::{vehicle_type::VehicleType, VehicleEnergyResult},
};

pub struct PHEV {
    pub name: String,
    pub charge_sustain_model: Arc<PredictionModelRecord>,
    pub charge_depleting_model: Arc<PredictionModelRecord>,
    pub battery_capacity: Energy,
    pub starting_battery_energy: Energy,
    pub battery_energy_unit: EnergyUnit,
    pub custom_liquid_fuel_to_kwh: Option<f64>,
}

impl PHEV {
    pub fn new(
        name: String,
        charge_sustain_model: PredictionModelRecord,
        charge_depleting_model: PredictionModelRecord,
        battery_capacity: Energy,
        starting_battery_energy: Energy,
        battery_energy_unit: EnergyUnit,
        custom_liquid_fuel_to_kwh: Option<f64>,
    ) -> Result<Self, TraversalModelError> {
        Ok(Self {
            name,
            charge_sustain_model: Arc::new(charge_sustain_model),
            charge_depleting_model: Arc::new(charge_depleting_model),
            battery_capacity,
            starting_battery_energy,
            battery_energy_unit,
            custom_liquid_fuel_to_kwh,
        })
    }
}

impl VehicleType for PHEV {
    fn name(&self) -> String {
        self.name.clone()
    }
    fn initial_state(&self) -> TraversalState {
        vec![
            StateVar(0.0),                                   // accumulated electrical energy
            StateVar(0.0),                                   // accumulated liquid_fuel energy
            StateVar(self.starting_battery_energy.as_f64()), // battery energy remaining
        ]
    }
    fn best_case_energy(
        &self,
        distance: (Distance, DistanceUnit),
    ) -> Result<(Energy, EnergyUnit), TraversalModelError> {
        let (distance, distance_unit) = distance;

        // assume lowest energy cost scenario for a PHEV is to just use the battery
        let energy = Energy::create(
            self.charge_depleting_model.ideal_energy_rate,
            self.charge_depleting_model.energy_rate_unit,
            distance,
            distance_unit,
        )?;
        Ok(energy)
    }
    fn consume_energy(
        &self,
        speed: (Speed, SpeedUnit),
        grade: (Grade, GradeUnit),
        distance: (Distance, DistanceUnit),
        state: &[StateVar],
    ) -> Result<VehicleEnergyResult, TraversalModelError> {
        let battery_soc_percentage = get_battery_soc_percent(self, state);

        let (
            electrical_energy,
            electrical_energy_unit,
            liquid_fuel_energy,
            liquid_fuel_energy_unit,
        ) = get_phev_energy(self, battery_soc_percentage, speed, grade, distance)?;

        // convert both energy sources to kWh
        let electrical_energy_kwh =
            electrical_energy_unit.convert(electrical_energy, EnergyUnit::KilowattHours);
        let liquid_fuel_energy_kwh = match self.custom_liquid_fuel_to_kwh {
            Some(custom_fuel_to_kwh) => {
                // use the custom conversion factor
                Energy::new(custom_fuel_to_kwh * liquid_fuel_energy.as_f64())
            }
            None => {
                // just use the default conversion factors
                liquid_fuel_energy_unit.convert(liquid_fuel_energy, EnergyUnit::KilowattHours)
            }
        };
        let total_energy_kwh = electrical_energy_kwh + liquid_fuel_energy_kwh;

        let updated_state = update_state(state, electrical_energy, liquid_fuel_energy);

        Ok(VehicleEnergyResult {
            energy: total_energy_kwh,
            energy_unit: EnergyUnit::KilowattHours,
            updated_state,
        })
    }
    fn serialize_state(&self, state: &[StateVar]) -> serde_json::Value {
        let battery_energy = get_electrical_energy_from_state(state);
        let liquid_fuel_energy = get_liquid_fuel_energy_from_state(state);
        let battery_soc_percent = get_battery_soc_percent(self, state);
        serde_json::json!({
            "battery_energy": battery_energy.as_f64(),
            "fuel_energy": liquid_fuel_energy.as_f64(),
            "battery_soc_percent": battery_soc_percent,
        })
    }

    fn serialize_state_info(&self, _state: &[StateVar]) -> serde_json::Value {
        let battery_energy_unit = self.battery_energy_unit;
        let fuel_energy_unit = self
            .charge_sustain_model
            .energy_rate_unit
            .associated_energy_unit();
        serde_json::json!({
            "battery_energy_unit": battery_energy_unit.to_string(),
            "fuel_energy_unit": fuel_energy_unit.to_string(),
        })
    }

    fn update_from_query(
        &self,
        query: &serde_json::Value,
    ) -> Result<Arc<dyn VehicleType>, TraversalModelError> {
        let starting_soc_percent = query
            .get("starting_soc_percent".to_string())
            .ok_or(TraversalModelError::BuildError(
                "No 'starting_soc_percent' key provided in query".to_string(),
            ))?
            .as_f64()
            .ok_or(TraversalModelError::BuildError(
                "Expected 'starting_soc_percent' value to be numeric".to_string(),
            ))?;
        if !(0.0..=100.0).contains(&starting_soc_percent) {
            return Err(TraversalModelError::BuildError(
                "Expected 'starting_soc_percent' value to be between 0 and 100".to_string(),
            ));
        }
        let starting_battery_energy = self.battery_capacity * (starting_soc_percent / 100.0);

        let new_phev = PHEV {
            name: self.name.clone(),
            charge_sustain_model: self.charge_sustain_model.clone(),
            charge_depleting_model: self.charge_depleting_model.clone(),
            battery_capacity: self.battery_capacity,
            starting_battery_energy,
            battery_energy_unit: self.battery_energy_unit,
            custom_liquid_fuel_to_kwh: self.custom_liquid_fuel_to_kwh,
        };

        Ok(Arc::new(new_phev))
    }
}

fn update_state(
    state: &[StateVar],
    electrical_energy: Energy,
    liquid_fuel_energy: Energy,
) -> TraversalState {
    let mut updated_state = Vec::with_capacity(state.len());

    // accumulated electrical energy
    updated_state.push(state[0] + electrical_energy.into());

    // accumulated fuel energy
    updated_state.push(state[1] + liquid_fuel_energy.into());

    // remaining battery energy
    let current_battery_energy = get_remaining_battery_energy_from_state(state);
    let new_battery_energy = (current_battery_energy - electrical_energy).max(Energy::new(0.0));
    updated_state.push(new_battery_energy.into());

    updated_state
}
fn get_electrical_energy_from_state(state: &[StateVar]) -> Energy {
    Energy::new(state[0].0)
}

fn get_liquid_fuel_energy_from_state(state: &[StateVar]) -> Energy {
    Energy::new(state[1].0)
}

fn get_remaining_battery_energy_from_state(state: &[StateVar]) -> Energy {
    Energy::new(state[2].0)
}

fn get_battery_soc_percent(vehicle: &PHEV, state: &[StateVar]) -> f64 {
    let battery_energy_unit = vehicle.battery_energy_unit;

    let battery_capacity_kwh =
        battery_energy_unit.convert(vehicle.battery_capacity, EnergyUnit::KilowattHours);

    let remaining_battery_energy = get_remaining_battery_energy_from_state(state);

    let remaining_battery_energy_kwh =
        battery_energy_unit.convert(remaining_battery_energy, EnergyUnit::KilowattHours);

    (remaining_battery_energy_kwh.as_f64() / battery_capacity_kwh.as_f64()) * 100.0
}

/// Compute the energy for the PHEV by converting liquid_fuel to kWh.
/// This uses a simplified operation in which we assume that if the battery
/// SOC is greater than zero we can just operate on battery to traverse a link.
/// This is not entirely realistic as it's possible to arrive at a link with
/// 0.001% SOC and still need to use liquid_fuel to traverse the link.
///
/// In the future we could make this more sophisticated by calculating
/// the energy required to traverse the link using the battery and then
/// finding the point at which we would have to switch to liquid_fuel
///
/// Returns a tuple of (electrical_energy, electrical_energy_unit, liquid_fuel_energy, liquid_fuel_energy_unit)
fn get_phev_energy(
    vehicle: &PHEV,
    battery_soc_percent: f64,
    speed: (Speed, SpeedUnit),
    grade: (Grade, GradeUnit),
    distance: (Distance, DistanceUnit),
) -> Result<(Energy, EnergyUnit, Energy, EnergyUnit), TraversalModelError> {
    let electrical_energy_unit = vehicle
        .charge_depleting_model
        .energy_rate_unit
        .associated_energy_unit();
    let liquid_fuel_energy_unit = vehicle
        .charge_sustain_model
        .energy_rate_unit
        .associated_energy_unit();

    if battery_soc_percent > 0.0 {
        // assume we can just use the battery
        let (electrical_energy, electrical_energy_unit) = vehicle
            .charge_depleting_model
            .predict(speed, grade, distance)?;
        Ok((
            electrical_energy,
            electrical_energy_unit,
            Energy::new(0.0),
            liquid_fuel_energy_unit,
        ))
    } else {
        // just use the liquid_fuel engine
        let (liquid_fuel_energy, liquid_fuel_energy_unit) = vehicle
            .charge_sustain_model
            .predict(speed, grade, distance)?;
        Ok((
            Energy::new(0.0),
            electrical_energy_unit,
            liquid_fuel_energy,
            liquid_fuel_energy_unit,
        ))
    }
}

#[cfg(test)]
mod tests {
    use routee_compass_core::util::unit::{EnergyRate, EnergyRateUnit};

    use crate::routee::{prediction::load_prediction_model, prediction::model_type::ModelType};

    use super::*;

    use std::path::PathBuf;

    fn mock_vehicle() -> PHEV {
        let charge_sustain_model_file_path = PathBuf::from(env!("CARGO_MANIFEST_DIR"))
            .join("src")
            .join("routee")
            .join("test")
            .join("2016_CHEVROLET_Volt_Charge_Sustaining.bin");
        let charge_depleting_model_file_path = PathBuf::from(env!("CARGO_MANIFEST_DIR"))
            .join("src")
            .join("routee")
            .join("test")
            .join("2016_CHEVROLET_Volt_Charge_Depleting.bin");

        let charge_sustain_model_record = load_prediction_model(
            "Chevy_Volt_Charge_Sustaining".to_string(),
            &charge_sustain_model_file_path,
            ModelType::Smartcore,
            SpeedUnit::MilesPerHour,
            GradeUnit::Decimal,
            EnergyRateUnit::GallonsGasolinePerMile,
            Some(EnergyRate::new(0.02)),
            Some(1.1252),
            None,
        )
        .unwrap();
        let charge_depleting_model_record = load_prediction_model(
            "Chevy_Volt_Charge_Depleting".to_string(),
            &charge_depleting_model_file_path,
            ModelType::Smartcore,
            SpeedUnit::MilesPerHour,
            GradeUnit::Decimal,
            EnergyRateUnit::KilowattHoursPerMile,
            Some(EnergyRate::new(0.2)),
            Some(1.3958),
            None,
        )
        .unwrap();

        PHEV::new(
            "Chevy_Volt".to_string(),
            charge_sustain_model_record,
            charge_depleting_model_record,
            Energy::new(12.0),
            Energy::new(12.0),
            EnergyUnit::KilowattHours,
            None,
        )
        .unwrap()
    }

    #[test]
    fn test_phev_energy_model_just_electric() {
        let vehicle = mock_vehicle();
        let initial = vehicle.initial_state();

        // starting at 100% SOC, we should be able to traverse 1000 meters
        // without using any liquid_fuel
        let distance = (Distance::new(1000.0), DistanceUnit::Meters);
        let speed = (Speed::new(60.0), SpeedUnit::MilesPerHour);
        let grade = (Grade::new(0.0), GradeUnit::Decimal);

        let result = vehicle
            .consume_energy(speed, grade, distance, &initial)
            .unwrap();

        let liquid_fuel_energy = get_liquid_fuel_energy_from_state(&result.updated_state);
        assert!(liquid_fuel_energy.as_f64() < 1e-9);

        let electrical_energy = get_electrical_energy_from_state(&result.updated_state);
        assert!(electrical_energy.as_f64() > 0.0);

        let battery_percent_soc = get_battery_soc_percent(&vehicle, &result.updated_state);
        assert!(battery_percent_soc < 100.0);
    }

    #[test]
    fn test_phev_energy_model_gas_and_electric() {
        let vehicle = mock_vehicle();
        let initial = vehicle.initial_state();

        // now let's traverse a really long link to deplete the battery
        let distance = (Distance::new(100.0), DistanceUnit::Miles);
        let speed = (Speed::new(60.0), SpeedUnit::MilesPerHour);
        let grade = (Grade::new(0.0), GradeUnit::Decimal);

        let result = vehicle
            .consume_energy(speed, grade, distance, &initial)
            .unwrap();

        let electrical_energy = get_electrical_energy_from_state(&result.updated_state);
        let battery_percent_soc = get_battery_soc_percent(&vehicle, &result.updated_state);

        assert!(electrical_energy.as_f64() > 0.0);
        assert!(battery_percent_soc < 1e-9);

        // and then traverse the same distance but this time we should only use liquid_fuel energy
        let result2 = vehicle
            .consume_energy(speed, grade, distance, &result.updated_state)
            .unwrap();

        let liquid_fuel_energy = get_liquid_fuel_energy_from_state(&result2.updated_state);

        assert!(liquid_fuel_energy.as_f64() > 0.0);
    }
}

==> nrel.routee.compass-0.6.0.tar.gz: nrel.routee.compass-0.6.0/rust/routee-compass-powertrain/src/routee/vehicle/default/phev.rs <==
use std::sync::Arc;

use routee_compass_core::{
    model::traversal::{
        state::{state_variable::StateVar, traversal_state::TraversalState},
        traversal_model_error::TraversalModelError,
    },
    model::unit::{
        as_f64::AsF64, Distance, DistanceUnit, Energy, EnergyUnit, Grade, GradeUnit, Speed,
        SpeedUnit,
    },
};

use crate::routee::{
    prediction::PredictionModelRecord,
    vehicle::{vehicle_type::VehicleType, VehicleEnergyResult},
};

pub struct PHEV {
    pub name: String,
    pub charge_sustain_model: Arc<PredictionModelRecord>,
    pub charge_depleting_model: Arc<PredictionModelRecord>,
    pub battery_capacity: Energy,
    pub starting_battery_energy: Energy,
    pub battery_energy_unit: EnergyUnit,
    pub custom_liquid_fuel_to_kwh: Option<f64>,
}

impl PHEV {
    pub fn new(
        name: String,
        charge_sustain_model: PredictionModelRecord,
        charge_depleting_model: PredictionModelRecord,
        battery_capacity: Energy,
        starting_battery_energy: Energy,
        battery_energy_unit: EnergyUnit,
        custom_liquid_fuel_to_kwh: Option<f64>,
    ) -> Result<Self, TraversalModelError> {
        Ok(Self {
            name,
            charge_sustain_model: Arc::new(charge_sustain_model),
            charge_depleting_model: Arc::new(charge_depleting_model),
            battery_capacity,
            starting_battery_energy,
            battery_energy_unit,
            custom_liquid_fuel_to_kwh,
        })
    }
}

impl VehicleType for PHEV {
    fn name(&self) -> String {
        self.name.clone()
    }

    fn state_variable_names(&self) -> Vec<String> {
        vec![
            String::from("energy_electric"),
            String::from("energy_liquid"),
            String::from("battery_state"),
        ]
    }

    fn initial_state(&self) -> TraversalState {
        vec![
            StateVar(0.0),                                   // accumulated electrical energy
            StateVar(0.0),                                   // accumulated liquid_fuel energy
            StateVar(self.starting_battery_energy.as_f64()), // battery energy remaining
        ]
    }
    fn best_case_energy(
        &self,
        distance: (Distance, DistanceUnit),
    ) -> Result<(Energy, EnergyUnit), TraversalModelError> {
        let (distance, distance_unit) = distance;

        // assume lowest energy cost scenario for a PHEV is to just use the battery
        let energy = Energy::create(
            self.charge_depleting_model.ideal_energy_rate,
            self.charge_depleting_model.energy_rate_unit,
            distance,
            distance_unit,
        )?;
        Ok(energy)
    }

    fn best_case_energy_state(
        &self,
        distance: (Distance, DistanceUnit),
        state: &[StateVar],
    ) -> Result<VehicleEnergyResult, TraversalModelError> {
        let (electrical_energy, electrical_energy_unit) = self.best_case_energy(distance)?;
        let updated_state = update_state(state, electrical_energy, Energy::ZERO);

        Ok(VehicleEnergyResult {
            energy: electrical_energy,
            energy_unit: electrical_energy_unit,
            updated_state,
        })
    }

    fn consume_energy(
        &self,
        speed: (Speed, SpeedUnit),
        grade: (Grade, GradeUnit),
        distance: (Distance, DistanceUnit),
        state: &[StateVar],
    ) -> Result<VehicleEnergyResult, TraversalModelError> {
        let battery_soc_percentage = get_battery_soc_percent(self, state);

        let (
            electrical_energy,
            electrical_energy_unit,
            liquid_fuel_energy,
            liquid_fuel_energy_unit,
        ) = get_phev_energy(self, battery_soc_percentage, speed, grade, distance)?;

        // convert both energy sources to kWh
        let electrical_energy_kwh =
            electrical_energy_unit.convert(electrical_energy, EnergyUnit::KilowattHours);
        let liquid_fuel_energy_kwh = match self.custom_liquid_fuel_to_kwh {
            Some(custom_fuel_to_kwh) => {
                // use the custom conversion factor
                Energy::new(custom_fuel_to_kwh * liquid_fuel_energy.as_f64())
            }
            None => {
                // just use the default conversion factors
                liquid_fuel_energy_unit.convert(liquid_fuel_energy, EnergyUnit::KilowattHours)
            }
        };
        let total_energy_kwh = electrical_energy_kwh + liquid_fuel_energy_kwh;

        let updated_state = update_state(state, electrical_energy, liquid_fuel_energy);

        Ok(VehicleEnergyResult {
            energy: total_energy_kwh,
            energy_unit: EnergyUnit::KilowattHours,
            updated_state,
        })
    }
    fn serialize_state(&self, state: &[StateVar]) -> serde_json::Value {
        let battery_energy = get_electrical_energy_from_state(state);
        let liquid_fuel_energy = get_liquid_fuel_energy_from_state(state);
        let battery_soc_percent = get_battery_soc_percent(self, state);
        serde_json::json!({
            "battery_energy": battery_energy.as_f64(),
            "fuel_energy": liquid_fuel_energy.as_f64(),
            "battery_soc_percent": battery_soc_percent,
        })
    }

    fn serialize_state_info(&self, _state: &[StateVar]) -> serde_json::Value {
        let battery_energy_unit = self.battery_energy_unit;
        let fuel_energy_unit = self
            .charge_sustain_model
            .energy_rate_unit
            .associated_energy_unit();
        serde_json::json!({
            "battery_energy_unit": battery_energy_unit.to_string(),
            "fuel_energy_unit": fuel_energy_unit.to_string(),
        })
    }

    fn update_from_query(
        &self,
        query: &serde_json::Value,
    ) -> Result<Arc<dyn VehicleType>, TraversalModelError> {
        let starting_soc_percent = query
            .get("starting_soc_percent".to_string())
            .ok_or_else(|| {
                TraversalModelError::BuildError(
                    "No 'starting_soc_percent' key provided in query".to_string(),
                )
            })?
            .as_f64()
            .ok_or_else(|| {
                TraversalModelError::BuildError(
                    "Expected 'starting_soc_percent' value to be numeric".to_string(),
                )
            })?;
        if !(0.0..=100.0).contains(&starting_soc_percent) {
            return Err(TraversalModelError::BuildError(
                "Expected 'starting_soc_percent' value to be between 0 and 100".to_string(),
            ));
        }
        let starting_battery_energy = self.battery_capacity * (starting_soc_percent / 100.0);

        let new_phev = PHEV {
            name: self.name.clone(),
            charge_sustain_model: self.charge_sustain_model.clone(),
            charge_depleting_model: self.charge_depleting_model.clone(),
            battery_capacity: self.battery_capacity,
            starting_battery_energy,
            battery_energy_unit: self.battery_energy_unit,
            custom_liquid_fuel_to_kwh: self.custom_liquid_fuel_to_kwh,
        };

        Ok(Arc::new(new_phev))
    }
}

fn update_state(
    state: &[StateVar],
    electrical_energy: Energy,
    liquid_fuel_energy: Energy,
) -> TraversalState {
    let mut updated_state = Vec::with_capacity(state.len());

    // accumulated electrical energy
    updated_state.push(state[0] + electrical_energy.into());

    // accumulated fuel energy
    updated_state.push(state[1] + liquid_fuel_energy.into());

    // remaining battery energy
    let current_battery_energy = get_remaining_battery_energy_from_state(state);
    let new_battery_energy = (current_battery_energy - electrical_energy).max(Energy::new(0.0));
    updated_state.push(new_battery_energy.into());

    updated_state
}
fn get_electrical_energy_from_state(state: &[StateVar]) -> Energy {
    Energy::new(state[0].0)
}

fn get_liquid_fuel_energy_from_state(state: &[StateVar]) -> Energy {
    Energy::new(state[1].0)
}

fn get_remaining_battery_energy_from_state(state: &[StateVar]) -> Energy {
    Energy::new(state[2].0)
}

fn get_battery_soc_percent(vehicle: &PHEV, state: &[StateVar]) -> f64 {
    let battery_energy_unit = vehicle.battery_energy_unit;

    let battery_capacity_kwh =
        battery_energy_unit.convert(vehicle.battery_capacity, EnergyUnit::KilowattHours);

    let remaining_battery_energy = get_remaining_battery_energy_from_state(state);

    let remaining_battery_energy_kwh =
        battery_energy_unit.convert(remaining_battery_energy, EnergyUnit::KilowattHours);

    (remaining_battery_energy_kwh.as_f64() / battery_capacity_kwh.as_f64()) * 100.0
}

/// Compute the energy for the PHEV by converting liquid_fuel to kWh.
/// This uses a simplified operation in which we assume that if the battery
/// SOC is greater than zero we can just operate on battery to traverse a link.
/// This is not entirely realistic as it's possible to arrive at a link with
/// 0.001% SOC and still need to use liquid_fuel to traverse the link.
///
/// In the future we could make this more sophisticated by calculating
/// the energy required to traverse the link using the battery and then
/// finding the point at which we would have to switch to liquid_fuel
///
/// Returns a tuple of (electrical_energy, electrical_energy_unit, liquid_fuel_energy, liquid_fuel_energy_unit)
fn get_phev_energy(
    vehicle: &PHEV,
    battery_soc_percent: f64,
    speed: (Speed, SpeedUnit),
    grade: (Grade, GradeUnit),
    distance: (Distance, DistanceUnit),
) -> Result<(Energy, EnergyUnit, Energy, EnergyUnit), TraversalModelError> {
    let electrical_energy_unit = vehicle
        .charge_depleting_model
        .energy_rate_unit
        .associated_energy_unit();
    let liquid_fuel_energy_unit = vehicle
        .charge_sustain_model
        .energy_rate_unit
        .associated_energy_unit();

    if battery_soc_percent > 0.0 {
        // assume we can just use the battery
        let (electrical_energy, electrical_energy_unit) = vehicle
            .charge_depleting_model
            .predict(speed, grade, distance)?;
        Ok((
            electrical_energy,
            electrical_energy_unit,
            Energy::new(0.0),
            liquid_fuel_energy_unit,
        ))
    } else {
        // just use the liquid_fuel engine
        let (liquid_fuel_energy, liquid_fuel_energy_unit) = vehicle
            .charge_sustain_model
            .predict(speed, grade, distance)?;
        Ok((
            Energy::new(0.0),
            electrical_energy_unit,
            liquid_fuel_energy,
            liquid_fuel_energy_unit,
        ))
    }
}

#[cfg(test)]
mod tests {
    use routee_compass_core::model::unit::{EnergyRate, EnergyRateUnit};

    use crate::routee::{prediction::load_prediction_model, prediction::model_type::ModelType};

    use super::*;

    use std::path::PathBuf;

    fn mock_vehicle() -> PHEV {
        let charge_sustain_model_file_path = PathBuf::from(env!("CARGO_MANIFEST_DIR"))
            .join("src")
            .join("routee")
            .join("test")
            .join("2016_CHEVROLET_Volt_Charge_Sustaining.bin");
        let charge_depleting_model_file_path = PathBuf::from(env!("CARGO_MANIFEST_DIR"))
            .join("src")
            .join("routee")
            .join("test")
            .join("2016_CHEVROLET_Volt_Charge_Depleting.bin");
        let model_type = ModelType::Interpolate {
            underlying_model_type: Box::new(ModelType::Smartcore),
            speed_lower_bound: Speed::new(0.0),
            speed_upper_bound: Speed::new(100.0),
            speed_bins: 101,
            grade_lower_bound: Grade::new(-0.20),
            grade_upper_bound: Grade::new(0.20),
            grade_bins: 41,
        };

        let charge_sustain_model_record = load_prediction_model(
            "Chevy_Volt_Charge_Sustaining".to_string(),
            &charge_sustain_model_file_path,
            model_type.clone(),
            SpeedUnit::MilesPerHour,
            GradeUnit::Decimal,
            EnergyRateUnit::GallonsGasolinePerMile,
            Some(EnergyRate::new(0.02)),
            Some(1.1252),
            None,
        )
        .unwrap();
        let charge_depleting_model_record = load_prediction_model(
            "Chevy_Volt_Charge_Depleting".to_string(),
            &charge_depleting_model_file_path,
            model_type.clone(),
            SpeedUnit::MilesPerHour,
            GradeUnit::Decimal,
            EnergyRateUnit::KilowattHoursPerMile,
            Some(EnergyRate::new(0.2)),
            Some(1.3958),
            None,
        )
        .unwrap();

        PHEV::new(
            "Chevy_Volt".to_string(),
            charge_sustain_model_record,
            charge_depleting_model_record,
            Energy::new(12.0),
            Energy::new(12.0),
            EnergyUnit::KilowattHours,
            None,
        )
        .unwrap()
    }

    #[test]
    fn test_phev_energy_model_just_electric() {
        let vehicle = mock_vehicle();
        let initial = vehicle.initial_state();

        // starting at 100% SOC, we should be able to traverse 1000 meters
        // without using any liquid_fuel
        let distance = (Distance::new(1000.0), DistanceUnit::Meters);
        let speed = (Speed::new(60.0), SpeedUnit::MilesPerHour);
        let grade = (Grade::new(0.0), GradeUnit::Decimal);

        let result = vehicle
            .consume_energy(speed, grade, distance, &initial)
            .unwrap();

        let liquid_fuel_energy = get_liquid_fuel_energy_from_state(&result.updated_state);
        assert!(liquid_fuel_energy.as_f64() < 1e-9);

        let electrical_energy = get_electrical_energy_from_state(&result.updated_state);
        assert!(electrical_energy.as_f64() > 0.0);

        let battery_percent_soc = get_battery_soc_percent(&vehicle, &result.updated_state);
        assert!(battery_percent_soc < 100.0);
    }

    #[test]
    fn test_phev_energy_model_gas_and_electric() {
        let vehicle = mock_vehicle();
        let initial = vehicle.initial_state();

        // now let's traverse a really long link to deplete the battery
        let distance = (Distance::new(100.0), DistanceUnit::Miles);
        let speed = (Speed::new(60.0), SpeedUnit::MilesPerHour);
        let grade = (Grade::new(0.0), GradeUnit::Decimal);

        let result = vehicle
            .consume_energy(speed, grade, distance, &initial)
            .unwrap();

        let electrical_energy = get_electrical_energy_from_state(&result.updated_state);
        let battery_percent_soc = get_battery_soc_percent(&vehicle, &result.updated_state);

        assert!(electrical_energy.as_f64() > 0.0);
        assert!(battery_percent_soc < 1e-9);

        // and then traverse the same distance but this time we should only use liquid_fuel energy
        let result2 = vehicle
            .consume_energy(speed, grade, distance, &result.updated_state)
            .unwrap();

        let liquid_fuel_energy = get_liquid_fuel_energy_from_state(&result2.updated_state);

        assert!(liquid_fuel_energy.as_f64() > 0.0);
    }
}
